""" All available blocks
"""
from ..entities import Block

blocks = (
    Block(block_id='fly', order=1),
    Block(block_id='shoot', order=2),
    Block(block_id='repeat', order=3),
    Block(block_id='while', order=4),
    Block(block_id='color', order=5),
    Block(block_id='position', order=6),
    Block(block_id='if', order=7),
    Block(block_id='if-else', order=8),
)
