""" All levels
"""
from ..entities import Level

# pylint:disable=line-too-long
levels = (
    Level(level_id=1, credits=0),
    Level(level_id=2, credits=4),
    Level(level_id=3, credits=9),
    Level(level_id=4, credits=16),
    Level(level_id=5, credits=25),
    Level(level_id=6, credits=36),
    Level(level_id=7, credits=49),
    Level(level_id=8, credits=64),
    Level(level_id=9, credits=81),
)
