""" Common entities for testing purposes
"""
from flocs.entities import Task, Student

s1 = Student(student_id=1, credits=0)

t1 = Task(task_id=1, category_id=None, setting=None, solution=None)
t2 = Task(task_id=2, category_id=None, setting=None, solution=None)
t3 = Task(task_id=3, category_id=None, setting=None, solution=None)
t4 = Task(task_id=4, category_id=None, setting=None, solution=None)
t5 = Task(task_id=5, category_id=None, setting=None, solution=None)
t6 = Task(task_id=6, category_id=None, setting=None, solution=None)
t7 = Task(task_id=7, category_id=None, setting=None, solution=None)
t8 = Task(task_id=8, category_id=None, setting=None, solution=None)
t9 = Task(task_id=9, category_id=None, setting=None, solution=None)
